@@ -3,9 +3,10 @@ import logging
 import os
 import platform
 import socket
+import time
 from dataclasses import dataclass
 from subprocess import Popen, PIPE, TimeoutExpired
-from typing import Any
+from typing import Any, Union
 
 from paramiko import SSHClient, ssh_exception, AutoAddPolicy
 
@@ -22,14 +23,14 @@ ch.setFormatter(formatter)
 logger.addHandler(ch)
 
 
-@dataclass()
+@dataclass
 class ResponseParser:
     """Bash response parser"""
 
     response: Any
 
     @property
-    def stdout(self) -> str:
+    def stdout(self) -> Union[str, dict]:
         return self.response[1]
 
     @property
@@ -38,7 +39,10 @@ class ResponseParser:
 
     @property
     def exited(self) -> int:
-        return int(self.response[0])
+        try:
+            return int(self.response[0])
+        except TypeError:
+            return 0
 
     @property
     def ok(self) -> bool:
@@ -48,6 +52,11 @@ class ResponseParser:
     def command(self) -> str:
         return self.response[3]
 
+    @property
+    def motd(self) -> str:
+        """Get message of the day"""
+        return self.response[4]
+
     def json(self):
         return json.loads(self.stdout)
 
@@ -67,7 +76,8 @@ class Plinux:
         self.port = port
         self.username = username
         self.password = password
-        logger.disabled = not logger_enabled
+        self.logger = logger
+        self.logger.disabled = not logger_enabled
 
     def __str__(self):
         return f'Local host: {self.get_current_os_name()}\n' \
@@ -91,8 +101,7 @@ class Plinux:
         """Returns all available public methods"""
         return [method for method in dir(self) if not method.startswith('_')]
 
-    @staticmethod
-    def run_cmd_local(cmd: str, timeout=60):
+    def run_cmd_local(self, cmd: str, timeout: int = 60):
         """Main function to send commands using subprocess
 
         :param cmd: string, command
@@ -104,18 +113,19 @@ class Plinux:
         with Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE) as process:
             # process.wait(timeout)
             try:
-                logger.info(f'COMMAND: "{cmd}"')
+                self.logger.info(f'COMMAND: "{cmd}"')
                 stdout, stderr = process.communicate(timeout=timeout)
                 data = (stdout + stderr).decode().strip()
-                logger.info(f'RESULT: "{data}"')
+                self.logger.info(f'RESULT: "{data}"')
                 return data
             except TimeoutExpired:
                 process.kill()
                 stdout, stderr = process.communicate()
                 data = stdout + stderr
+                self.logger.exception('Connection timeout')
                 return data.decode()
 
-    def _client(self, sftp=False, timeout=15):
+    def _client(self, sftp: bool = False, timeout: int = 15):
         """http://www.paramiko.org/"""
 
         client = SSHClient()
@@ -127,15 +137,15 @@ class Plinux:
             if sftp:
                 return client.open_sftp()
             return client
-        except ssh_exception.AuthenticationException as e:
-            logger.error(e.args)
-            raise ssh_exception.AuthenticationException
-        except ssh_exception.NoValidConnectionsError as e:
-            logger.error(e.strerror)
-            raise ssh_exception.NoValidConnectionsError
-        except TimeoutError as e:
-            logger.error('Timeout exceeded.' + e.strerror)
-            raise TimeoutError
+        except ssh_exception.AuthenticationException as err:
+            self.logger.exception('Authentication error')
+            raise err
+        except ssh_exception.NoValidConnectionsError as err:
+            self.logger.exception('There is no valid connection')
+            raise err
+        except TimeoutError as err:
+            self.logger.exception('Timeout exceeded.' + err.strerror)
+            raise err
 
     def run_cmd(self, cmd: str, sudo: bool = False, timeout: int = 30) -> ResponseParser:
         """Base method to execute SSH command on remote server
@@ -150,7 +160,7 @@ class Plinux:
 
         try:
             command = f"sudo -S -p '' -- sh -c '{cmd}'" if sudo else cmd
-            logger.info(command)
+            self.logger.info(f'[{self.host}] ' + command)
 
             stdin, stdout, stderr = client.exec_command(command, timeout=timeout)
 
@@ -164,7 +174,7 @@ class Plinux:
             # Get STDOUT
             stdout = stdout.read().decode().strip()
             out = stdout if stdout else None
-            logger.info(f'{exited}: {out}')
+            self.logger.info(f'{exited}: {out}')
 
             # Get STDERR
             stderr = stderr.read().decode().strip()
@@ -175,13 +185,46 @@ class Plinux:
 
             err = stderr if stderr else None
             if err:
-                logger.error(err)
+                self.logger.error(err)
 
             response = exited, out, err, command
             return ResponseParser(response)
         finally:
             client.close()
 
+    # noinspection PyTypeChecker
+    def run_cmd_session(
+            self,
+            cmds: tuple = None,
+            buffer_size: int = 4096,
+    ) -> ResponseParser:
+        """Send command using session. Returns entity name if noe command provided.
+
+        :param cmds: List of commands to execute
+        :param buffer_size: Bytes size to receive. Default 4096.
+        :return:
+        """
+
+        client = self._client()
+
+        with client.invoke_shell() as ssh:
+            self.logger.info(f'[{self.host}] ' + ' -> '.join(cmds))
+
+            ssh.settimeout(5)
+            motd = ssh.recv(buffer_size).decode().strip()
+
+            result = {}
+
+            # Send command one-by-one
+            for cmd in cmds:
+                ssh.send(f'{cmd}\n')
+                time.sleep(0.5)
+                response = ssh.recv(buffer_size).decode().splitlines()
+                result[cmd] = response[1:-1]
+
+            response = None, result, result, cmds, motd
+            return ResponseParser(response)
+
     @staticmethod
     def get_current_os_name():
         return platform.system()
@@ -195,6 +238,7 @@ class Plinux:
             self.run_cmd('whoami')
             return True
         except ssh_exception.AuthenticationException:
+            self.logger.exception('Invalid credentials')
             return False
 
     def get_os_version(self):
@@ -227,8 +271,8 @@ class Plinux:
         """Get service status"""
         return self.run_cmd(f'systemctl is-active {name}')
 
-    def is_service_active(self, name: str):
-        return True if self.get_service_status(name) == 'active' else False
+    def is_service_active(self, name: str) -> bool:
+        return True if self.get_service_status(name).stdout == 'active' else False
 
     def stop_service(self, name: str):
         return self.run_cmd(f'systemctl stop {name}', sudo=True)
@@ -280,6 +324,7 @@ class Plinux:
         try:
             return int(result)
         except TypeError as err:
+            self.logger.exception(f'Cannot get pid ({name})')
             raise err
 
     def get_netstat_info(self, params: str = ''):
@@ -332,32 +377,35 @@ class Plinux:
             print(json.dumps(jsoned, indent=4), sep='')
         return jsoned
 
-    def create_file(self, path: str):
-        return self.run_cmd(f'touch {path}', sudo=True)
+    def create_file(self, path: str, sudo: bool = True):
+        return self.run_cmd(f'touch {path}', sudo=sudo)
 
-    def get_file_permissions(self, path: str, human: bool = False):
+    def get_file_permissions(self, path: str, human: bool = False, sudo: bool = False):
         """Display file or file system status.
 
         :param path: File path
         :param human: Access rights in human readable form otherwise in in octal
+        :param sudo:
         :return:
         """
 
         cmd = f'stat -c %A {path}' if human else f'stat -c %a {path}'
-        return self.run_cmd(cmd)
+        return self.run_cmd(cmd, sudo=sudo)
 
-    def get_file_size(self, path: str):
+    def get_file_size(self, path: str, sudo: bool = False):
         """Get file size
 
+        :param sudo:
         :param path: File path
         :return: size in bytes
         """
 
-        return self.run_cmd(f'stat -c "%s" {path}')
+        return self.run_cmd(f'stat -c "%s" {path}', sudo=sudo)
 
-    def grep_line_in_file(self, path: str, string: str, directory: bool = False):
+    def grep_line_in_file(self, path: str, string: str, directory: bool = False, sudo: bool = True):
         """Grep line in file or directory
 
+        :param sudo:
         :param path: File/directory path
         :param string: string pattern to grep
         :param directory: If True - grep in directory with files
@@ -365,26 +413,28 @@ class Plinux:
         """
 
         if directory:
-            return self.run_cmd(f'grep -rn "{string}" {path}')
-        return self.run_cmd(f'grep -n "{string}" {path}')
+            return self.run_cmd(f'grep -rn "{string}" {path}', sudo=sudo)
+        return self.run_cmd(f'grep -n "{string}" {path}', sudo=sudo)
 
-    def change_line_in_file(self, path: str, old: str, new: str):
+    def change_line_in_file(self, path: str, old: str, new: str, sudo: bool = True):
         """Replace line and save file.
 
+        :param sudo:
         :param path: File path
         :param old: String to replace
         :param new: New string
         :return:
         """
 
-        return self.run_cmd(f'sed -i "s!{old}!{new}!" {path}', sudo=True)
+        return self.run_cmd(f'sed -i "s!{old}!{new}!" {path}', sudo=sudo)
 
-    def delete_line_from_file(self, path: str, string: str):
-        return self.run_cmd(f"sed -i '/{string}/d' {path}", sudo=True)
+    def delete_line_from_file(self, path: str, string: str, sudo: bool = True):
+        return self.run_cmd(f"sed -i '/{string}/d' {path}", sudo=sudo)
 
-    def get_last_file(self, directory: str = '', name: str = ''):
+    def get_last_file(self, directory: str = '', name: str = '', sudo: bool = True):
         """Get last modified file in a directory
 
+        :param sudo:
         :param name: Filename to grep
         :param directory: Directory path to precess. Home by default
         :return:
@@ -392,9 +442,9 @@ class Plinux:
 
         directory_ = directory or f'/home/{self.username}'
         cmd = f'ls {directory_} -Art| grep {name} | tail -n 1' if name else f'ls {directory} -Art | tail -n 1'
-        return self.run_cmd(cmd)
+        return self.run_cmd(cmd, sudo=sudo)
 
-    def remove(self, path: str):
+    def remove(self, path: str, sudo: bool = True):
         """Remove file(s) and directories
 
         Usage:\n
@@ -402,14 +452,16 @@ class Plinux:
         path=/opt/1/* remove all file in the directory\n
         path=/opt/1/file.txt remove specified file in the directory\n
 
+        :param sudo:
         :param path: Path to a file or directory.
         """
 
-        return self.run_cmd(f'for file in {path}; do rm -rf "$file"; done', sudo=True)
+        return self.run_cmd(f'for file in {path}; do rm -rf "$file"; done', sudo=sudo)
 
-    def extract_files(self, src: str, dst: str, mode: str = 'tar', quite: bool = True):
+    def extract_files(self, src: str, dst: str, mode: str = 'tar', quite: bool = True, sudo: bool = False):
         """Extract file to specified directory
 
+        :param sudo:
         :param src: Full path to archive (with extension)
         :param dst:
         :param mode: "tar", "zip"
@@ -422,27 +474,29 @@ class Plinux:
 
         cmd = tar_cmd if mode == 'tar' else unzip_cmd
 
-        return self.run_cmd(cmd)
+        return self.run_cmd(cmd, sudo=sudo)
 
-    def copy_file(self, src: str, dst: str):
+    def copy_file(self, src: str, dst: str, sudo: bool = True):
         """Copy file to another location
 
+        :param sudo:
         :param src: Source full path
         :param dst: Destination
         :return:
         """
 
-        return self.run_cmd(f'cp {src} {dst}', sudo=True)
+        return self.run_cmd(f'cp {src} {dst}', sudo=sudo)
 
-    def get_md5(self, path: str, raw: bool = False):
+    def get_md5(self, path: str, raw: bool = False, sudo: bool = True):
         """
         Get file md5
 
+        :param sudo:
         :param path: File path
-        :param raw: Return md5 sum only (not Rseponse object)
+        :param raw: Return md5 sum only (not Response object)
         """
 
-        result = self.run_cmd(f'md5sum {path}').stdout
+        result = self.run_cmd(f'md5sum {path}', sudo=sudo).stdout
         if raw:
             return result.split(path)[0].strip()
         return result
@@ -455,22 +509,23 @@ class Plinux:
         return self.run_cmd('shutdown -r now', sudo=True)
 
     def shutdown(self):
-        return self.run_cmd('shutdown -h now', True)
+        return self.run_cmd('shutdown -h now', sudo=True)
 
     #  ----------- Directory management -----------
-    def create_directory(self, path: str):
-        return self.run_cmd(f'mkdir {path}', sudo=True)
+    def create_directory(self, path: str, sudo: bool = True):
+        return self.run_cmd(f'mkdir {path}', sudo=sudo)
 
-    def list_dir(self, path: str, params=None):
+    def list_dir(self, path: str, params=None, sudo: bool = False):
         """List directory
 
         :param path: Directory path
         :param params: Additional params. For example: "la"
+        :param sudo:
         :return:
         """
 
         cmd = f'ls {path} -{params}' if params else f'ls {path}'
-        return self.run_cmd(cmd)
+        return self.run_cmd(cmd, sudo=sudo)
 
     def count_files(self, path: str):
         return self.run_cmd(f'ls {path} | wc -l')
@@ -491,7 +546,7 @@ class Plinux:
         """
 
         self.sftp.put(local, remote, confirm=True)
-        logger.info(f'Uploaded {local} to {remote}')
+        self.logger.info(f'Uploaded {local} to {remote}')
         return self.exists(remote)
 
     def download(self, remote: str, local: str, callback=None) -> bool:
@@ -506,7 +561,7 @@ class Plinux:
         """
 
         self.sftp.get(remote, local, callback=callback)
-        logger.info(f'Downloaded {remote} to {local}')
+        self.logger.info(f'Downloaded {remote} to {local}')
         return self.exists(local)
 
     def change_password(self, new_password: str):
@@ -532,13 +587,13 @@ class Plinux:
     def debug_info(self):
         """Show debug log. Logger must be enabled"""
 
-        logger.info('Linux client created.')
-        logger.info(f'Local host: {self.get_current_os_name()}')
-        logger.info(f'Remote IP: {self.host}')
-        logger.info(f'Username: {self.username}')
-        logger.info(f'Password: {self.password}')
-        logger.info(f'Available: {self.is_host_available()}')
-        logger.info(f'Available: {self.is_host_available()}')
+        self.logger.info('Linux client created.')
+        self.logger.info(f'Local host: {self.get_current_os_name()}')
+        self.logger.info(f'Remote IP: {self.host}')
+        self.logger.info(f'Username: {self.username}')
+        self.logger.info(f'Password: {self.password}')
+        self.logger.info(f'Available: {self.is_host_available()}')
+        self.logger.info(f'Available: {self.is_host_available()}')
 
     # ---------- User management ----------
     def kill_user_session(self, name: str):
